// Copyright 2011 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.api.ads.dfa.lib.factory;

import com.google.api.ads.common.lib.factory.FactoryConfiguration;
import com.google.api.ads.dfa.lib.DfaModule;
import com.google.api.ads.dfa.lib.client.DfaSession;
import com.google.common.annotations.VisibleForTesting;

/**
 * Utility class to create DFA service clients. The session is stored
 * as a reference within the client so that the session may be altered to
 * affect the behavior of the client.
 * <p>
 * Usage example:
 * <pre><code>
 * DfaSession session = new DfaSession.Builder()...build();
 * CampaignRemote campaignService =
 *     new DfaServices().get(session, CampaignRemote.class);
 * campaignService.makeRequest(...);
 * ...
 * session.setToken(...);
 * campaignService.makeRequest(...);
 * </code></pre>
 *
 * @author Joseph DiLallo
 */
public final class DfaServices {

  private DfaServiceClientFactory factory;

  /**
   * Default constructor.
   */
  public DfaServices() {
    factory = new DfaServiceClientFactory();
  }

  /**
   * Constructor that takes a factory.
   * @param factory the factory to create services from.
   */
  @VisibleForTesting
  DfaServices(DfaServiceClientFactory factory) {
    this.factory = factory;
  }

  /**
   * Constructor that takes a factory configuration.
   * @param factoryConfiguration the factory configuration
   */
  public DfaServices(FactoryConfiguration factoryConfiguration) {
    this.factory = new DfaServiceClientFactory(new DfaModule(factoryConfiguration));
  }

  /**
   * Gets a client for the service represented by the interface with a reference
   * to the session.
   *
   * @param <T> the service type
   * @param dfaSession the DFA session
   * @param interfaceClass the service interface class
   * @return the client for the service
   */
  public <T> T get(DfaSession dfaSession,
      Class<T> interfaceClass) {
    return factory.getServiceClientAsInterface(dfaSession, interfaceClass);
  }
}
